; @(#)qp-sample.el	1.3 9/21/94
; Author: Tom Howland (tom@rahul.net)

; do some key bindings and font manipulations.

; first, you must modify your load path. It should go something like this:

; (setq load-path (append '("/opt/quintus/editor/gnu3.2") load-path))

; then load this library file

; (load-library "qp")

; of course, it is up to the installer where quintus is actually
; installed, so /opt may or may not be the right place to start.

(autoload 'prolog-mode "qprolog-mode")
(autoload 'run-prolog "qprolog-mode" "Run an inferior prolog process" t)

; Some versions of cperl-mode manipulate auto-mode-alist in a way that
; conflicts with prolog-mode. The file suffix '.pl' is claimed by both
; Prolog and Perl.
;
; consider putting
;
;	# -*- mode: Perl; -*-
;
; as the first line in your Perl source files, or maybe delete the
; following and put
;
;	% -*- mode: Prolog; -*-
;
; as the first line in your Prolog source files.

(condition-case nil (require 'cperl-mode) (error t))

; remrassq is built-in Xemacs

(or (fboundp 'remrassq)
  (defun remrassq (a l)

"The proper definition:

(remrassq VALUE LIST)

Delete by side effect any elements of LIST whose cdr is `eq' to VALUE.
The modified LIST is returned.  If the first member of LIST has a car
that is `eq' to VALUE, there is no way to remove it by side effect;
therefore, write `(setq foo (remrassq value foo))' to be sure of changing
the value of `foo'.

This version is emphatically not based on side effects."

    (if l
	(let ((h (car l)) (r (cdr l)))
	  (if (eq a (cdr h)) (remrassq a r)
	    (cons h (remrassq a r)))))))

(setq auto-mode-alist
      (cons
       '("\\(\\.\\(pl\\|prolog\\|m\\)\\|\\`\\(prolog\\.ini\\|\\.\\(pl\\|sicstus\\)rc\\)\\)\\'" . prolog-mode)
       (remrassq 'prolog-mode auto-mode-alist)))

; we want Prolog compilation errors to be just as selectable as any
; other error in a compilation window

(defvar compilation-mode-hook-been-here nil)
(defun qp-compilation-mode-hook ()
  (if compilation-mode-hook-been-here t
    (setq compilation-error-regexp-alist
	  (append
	   '(("[*!] Approximate line: \\([0-9]+\\), file: '\\(.*\\)'" 2 1)
	     ("[*!] between lines \\([0-9]+\\) and [0-9]+ in file \\(.*\\)$" 2 1))
	   compilation-error-regexp-alist))
    (setq compilation-mode-hook-been-here t)))
(add-hook 'compilation-mode-hook 'qp-compilation-mode-hook)

; provide the capability to run different prologs. That is, different
; development systems with statically linked code (see qld -D for more info)

(defun really-run-prolog (arg)
  "Allows you to run a prolog with a specific path. If your PATH is set correctly, you only need to specify the executables name. For example, just entering \"prolog\" should work."
  (interactive "FWhich Prolog? ")
  (setenv "QUINTUS_PROLOG_PATH" arg)
  (run-prolog))

(defun really-run-prolog-with-args (arg args)
  (interactive "FWhich Prolog? 
sWith args: ")
  (load "qprolog-mode" nil t)
  (setenv "QUINTUS_PROLOG_PATH" arg)
  (setq startup-jcl (concat " +C" " Emacs:" prolog-zap-file " " args))
  (run-prolog))

(defun run-last-prolog ()
  (interactive)
  (message
   (or (getenv "QUINTUS_PROLOG_PATH")
       "default prolog"))
  (run-prolog))

(define-key ctl-x-map "p" 'really-run-prolog)
(define-key ctl-x-map "R" 'run-last-prolog)
(define-key ctl-x-map "r" 'really-run-prolog-with-args)

;; Xemacs tricks ... highlighting of strings and comments in prolog-mode

(cond
 ((string-match "XEmacs" emacs-version)
  (add-hook 'prolog-mode-hook   'turn-on-font-lock)
  (add-hook 'comint-prolog-hook 'turn-on-font-lock)))
